/*
    MSGLIB - a message handling library
    Copyright (C) 1995,1997  Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    See: COPYING.LB
*/
/*
    MSGLIB - a message handling library
    Copyright (C) 1995,1997  Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    See: COPYING.LB
*/
/* $RCSfile: MSG2LOCK.C $
   $Locker:  $	$Name:  $	$State: Exp $

	msgLock()/msgUnlock() functions for method #2.

	The method #2 to access messages bases upon message stored within a
	file. There may be stored multiple languages within the file, but
	the file may contain data other than message definitions.

	When a language is searched for a language code must be supplied.
	This code is identical to DOS's country code.

	There can be supplied more than one code, the codes will be searched
	in the specified order. If no match is found, the codes are searched
	in the alias lists. If also no match is found, the very first
	language is used. Because of this, the first language should be
	English and English should not be included into the alias list
	unless English is very close to the language.

	There is currently only one file format. This is because to make
	the retrieval process simple and, I suppose, to optimize the search
	for a country is not necessary as it will be done only once, normally.
	Also, the addresses are relatively to ease the way to append language
	files together and to store other information there as well.
	Structure #0 is required for some functions only.

	The messages are identified by one group and one number. When a
	language is choosen, the address to structure #2 is located and
	cached. When a message is locked, the following steps are performed:
		1) locate the message in the heap
		2) if failed: locate the structure #2 of the particular message group
		3) if successful: locate message starting at structure #3
		4) if successful: duplicate the message into the heap

	structure #0 "id string"
	Offset	Length	Meaning
	-11		DWord	address of structure #1 relative to Offset +0
	-7		Byte	Version id: '0' -> linked list with aliases
	-6		6 Bytes	"MSGLIB"	id string
	+0		<end of structure>

	<linked list with aliases>
	structure #1 "language table"
	Offset	Length	Meaning
	+0		Word	code of language
	+2		DWord	address of next lng relative to Offset +0, 0 if no next
	+6		Byte	group id of the messages
	+7		Word	number of aliases, may be 0
	+9		Word	alias code(s) of this language
	N*2+9	<structure #2>

	Note: The DWord at offset +2 may point to a location outside the
	file, this is detected and is considered to be "end of list" as
	well. This can be used to construct message libraries this way:
	Generate one binary file per language, place structure #1 at offset
	0 of the file and fill the complete length of the file into offset
	+2 of structure #1.  Now, any number of binary files can be concated
	to a single file, and this file can be used as a message library!

	structure #2 "message group"
	Offset	Length	Meaning
	+0		Word	number of defined messages of group, never 0
	+2		DWord	address(es) of messages relative to Offset +0
	N*4+2	<end of structure>

	structure #3 "message definition"
	Offset	Length	Meaning
	+0		Byte	Exit code
	+1		Word	Length of message string, may be 0
	+3		Byte	first character of the message string
	N+3		<end of structure>

	heap entry:
	struct MsgHeapEntry {
		struct MsgHeapEntry *nxt;
		MSGID msg;
		word linkcount;
		byte exitcode;
		char string[1];
	};
	The size of the entry grows automatically to fit the message string
	into the string[] array.
	If the same message is locked twice, only the linkcount is increased
	and the address of string[] is returned.  When a message is
	unlocked, the linkcount is decremented, if it reaches zero, the
	entry is destroyed.

*/


#ifndef lint
static char const rcsid[] = 
	"$Id: MSG2LOCK.C 1.2 1998/01/14 05:52:38 ska Exp $";
#endif

#include <stdlib.h>
#include <portable.h>
#include "mcvararg.h"
#include "yerror.h"
#include "msgconf2.h"

struct msgGroupEntry {			/* where a particular group starts */
	MSGFILE mge_file;			/* the file handle/pointer */
	word mge_maxnum;			/* number of message in this group */
	dword mge_position;			/* start of structure #2 offset +2 */
};


/*
 *	Currently there is a maximum of 16 message groups, that makes
 *	a total of 8*16=128 bytes. As currently only two groups are assigned
 *	that would waste 112 bytes in the data segment, but it will spare
 *	bytes within the code segment as this is a simple array now.
 */
static struct msgGroupEntry group[MSGGROUPMAX];

/*
 *	Here the chain of all locked messages starts
 */
static struct msgHeapEntry *head = NULL;

/*
 *	previous element of the last search
 */
static struct msgHeapEntry *prev = NULL;


/******************************
				file functions
 										***************/

#ifndef _MICROC_
#define Mseekd(f,pos) (lseek(f, pos, 0) == pos)
#else
int Mseekd(MSGFILE f, dword *pos)
{	return lseek(f, pos->hi, pos->lo, 0) == 0;
}
#endif

/*
 *	Seek to an entry of structure #2
 *	num cannot exceed 0xFFF, therefore it is completely save to shift
 *	the num within the (word) value range.
 */
int Mseek(MSGFILE f, dword iM(*)pos, int num)
{	dword p;

	longset(p, num << 2);
	longadd(p, pos);
	return Mseekd(f, p);
}

#define Mread(f,buf,len) (read(buf, len, f) == len)

/******************************
				base functions
 										***************/

/*
 *	Initialize the message retrieval system
 */
VARIANT(msg2Init, MSGFILE f)
{	
	VARIANTINIT(ap, f)

	if(f == MSGNOFILE) {


/*
 *	search for a message and preserve the previous element
 */
static struct MsgHeapEntry *searchChain(MSGID msg)
{	struct MsgHeapEntry *act;

	prev = NULL;
	if((act = head) != NULL) do {
		if(act->mhe_msg == msg)
			return act;
		prev = act;
	} while((act = act->nxt) != NULL);

	return NULL;
}

/*
 *	Initialize a heap entry and chain it into the list
 */
void initEntry(struct MsgHeapEntry *act, MSGID msg, int errnr)
{
	act->mhe_exitcode = errnr;
	act->mhe_msg = msg;
	act->mhe_nxt = head;
	act->mhe_linkcount = 1;
	head = act;
}

/*
 *	generate a default entry, if the message id is invalid or
 *	an error occured
 *	The default entry consists of: "message #<msg>\n"
 *	If no memory is availabe, "fatal(E_noMem)" is invoked.
 */
struct MsgHeapEntry *mkDfltEntry(MSGID msg)
{	struct MsgHeapEntry *h;

	if((h = malloc(sizeof(struct MsgHeapEntry) + 6 + strlen(Y_dlftMsg))) == NULL)
		fatal(E_noMem);
	sprintf(&h->mhe_string, Y_dlftMsg, msg);
	initEntry(h, msg, 255);

	return h;
}

/*
 *	Lock a message, thus, duplicate it into the heap
 */
char *msg2Lock(MSGID msg)
{	struct MsgHeapEntry *act;
	MSGFILE f;
	struct MsgGroupEntry *pGroup;
	struct MsgStruct3 ms3;

	if((act = searchChain(msg)) == NULL) {	/* not in heap -> load from file */
		if((f = (pGroup = &group[msgGroup(msg)])->mge_file) == MSGNOFILE
		 || msgId(msg) >= pGroup->mge_maxnum
		 || !Mseek(f, pGroup->mge_position, msgId(msg) - 1)
		 || !Mread(f, aS(ms3), sizeof(ms3))
		 || !(act = malloc(sizeof(struct MsgHeapEntry) + ms3.ms3_length)))
			/* message not found -> construct default entry */
			act = mkDfltmsg(msg);
		else if(!Mread(f, &(act->mhe_string), ms3.ms3_length)) {
			free(act);
			act = mkDfltmsg(msg);
		}
		else {		/* string read, now initialize the other fields */
			(act->mhe_string)[ms3.ms3_length] = NUL;
			initEntry(act, msg, ms3.ms3_exitcode);
		}
	}
	else ++act->mhe_linkcount;

	return &(act->mhe_string);
}

/*
 *	Unlock a message in the heap
 */
void msg2Unlock(MSGID msg)
{	struct MsgHeapEntry *act;

	if((act = searchChain(msg)) != NULL		/* OK, found -> unlock */
	 && --act->mhe_linkcount == 0) {		/* no further link -> unlink */
	 	if(!prev) head = act->mhe_nxt;
	 	else prev->mhe_nxt = act->mhe_nxt;
	 	free(act);
	}
}

/*
 *	Unlock a message via its message string
 */
void msg2UnlockString(char *s)
{	struct MsgHeapEntry *act;

	if((act = head) != NULL) do {
		if(&(act->mhe_string) == s) {
			msg2Unlock(act->mhe_msg);
			return;
		}
	} while((act = act->mhe_nxt) != NULL);
}

/*
 *	Retrieve a message and automatically lock/unlock 
 */
char *msg2Retrieve(MSGID msg)
{	static MSGID last = 0;

	msg2Unlock(last);
	return msg2Lock(last = msg);
}
